import { MANUAL_OPEN_GAME, RESET_STEP, SCORE_GAME_STEP } from 'common/Constants'
import { ActivityGame } from 'models/ActivityGame'
import { DAILY_SIGN, SIGN_TOTAL, TicketRecord, USE_TICKET } from 'models/TicketRecord'
import { queryCheckInList } from 'services/chain.svr'
import { checkInToday, seqSignCfg, seqSignScore, totalSignCfg, totalSignScore } from 'services/sign.svr'
import { ZError, SyncLocker, BaseController, router } from 'zutils'
import { formatDate } from 'zutils/utils/date.util'
import { generateChestLevel, generateNewChest, generateStepReward } from 'services/game.svr'
import { ChestStatusEnum } from 'models/ActivityChest'
import { updateRankScore } from 'services/rank.svr'

/**
 * 探索游戏相关接口
 */
class GameController extends BaseController {
  /**
   * 签到
   */
  @router('post /api/user/checkin')
  async checkIn(req) {
    new SyncLocker().checkLock(req)
    const user = req.user
    const { address } = user
    const dateTag = formatDate(new Date())
    const gameRecord = await ActivityGame.insertOrUpdate({ user: user.id, activity: user.activity }, {})
    if (MANUAL_OPEN_GAME && gameRecord.status === 0) {
      throw new ZError(11, 'map not open')
    }
    if (dateTag === gameRecord.lastSignDay) {
      throw new ZError(12, 'already claimed')
    }
    const record = await checkInToday(address, dateTag)
    if (!record) {
      throw new ZError(13, 'had not signed in')
    }
    const reward = seqSignScore(record.count)
    const ticketRecord = new TicketRecord({
      user: user.id,
      activity: user.activity,
      type: DAILY_SIGN,
      data: {},
      score: 1 + reward,
    })
    await ActivityGame.updateOne(
      { user: user.id, activity: user.activity },
      { lastSignDay: dateTag, $inc: { tickets: 1 + reward } },
    )
    await ticketRecord.save()
    return { ticket: reward + 1 }
  }

  /**
   * 签到列表
   */
  @router('get /api/user/checkin/list/:tag')
  async checkInList(req) {
    const user = req.user
    const { tag } = req.params
    let days: any = 1
    if (tag === '1month') {
      days = '1month'
    } else if (tag === 'last') {
      days = '1'
    }
    const res = await queryCheckInList(user.address.toLowerCase(), days, 0)
    return res
  }
  /**
   * 领取累计签到奖励
   */
  @router('post /api/user/checkin/claim')
  async claimCheckResult(req) {
    new SyncLocker().checkLock(req)
    const user = req.user
    let { days } = req.params
    if (!days || isNaN(days)) {
      throw new ZError(11, 'invalid days')
    }
    days = parseInt(days)
    if (days < 1) {
      throw new ZError(12, 'invalid days')
    }
    const dateTag = formatDate(new Date())
    const checkRecord = await checkInToday(user.address.toLowerCase(), dateTag)
    if (!checkRecord) {
      throw new ZError(12, 'not signed in')
    }
    if (days > checkRecord.total) {
      throw new ZError(13, 'invalid days')
    }
    const ticketRecords = await TicketRecord.find({ user: user.id, activity: user.activity, type: SIGN_TOTAL })
    const claimedSet = new Set()
    ticketRecords.forEach(record => {
      claimedSet.add(record.data.day)
    })
    if (claimedSet.has(days)) {
      throw new ZError(14, 'already claimed')
    }

    const score = totalSignScore(days)
    if (score === 0) {
      throw new ZError(15, 'invalid days')
    }
    const ticketRecord = new TicketRecord({
      user: user.id,
      activity: user.activity,
      type: SIGN_TOTAL,
      data: { day: days },
      score,
    })
    const gameRecord = await ActivityGame.insertOrUpdate({ user: user.id, activity: user.activity }, {})
    if (MANUAL_OPEN_GAME && gameRecord.status === 0) {
      throw new ZError(12, 'map not open')
    }
    await ActivityGame.updateOne({ user: user.id, activity: user.activity }, { $inc: { tickets: score } })
    await ticketRecord.save()
    return { ticket: score }
  }
  /**
   * 开启地图
   */
  @router('get /api/game/open')
  async openMap(req) {
    const user = req.user
    const gameInfo = await ActivityGame.insertOrUpdate({ user: user.id, activity: user.activity }, {})
    if (gameInfo.status === 1) {
      throw new ZError(11, 'already opened')
    }
    // check if all task finished
    if (!user.allTaskFinished()) {
      throw new ZError(12, 'The opening conditions are not met')
    }
    gameInfo.status = 1
    await gameInfo.save()
    return {}
  }
  /**
   * 探索状态
   */
  @router('get /api/game/stat')
  async gameStat(req, res) {
    const user = req.user
    const record = await ActivityGame.insertOrUpdate({ user: user.id, activity: user.activity }, {})
    if (MANUAL_OPEN_GAME && record.status === 0) {
      throw new ZError(12, 'map not open')
    }
    const signCfg = seqSignCfg
    const dateTag = formatDate(new Date())
    const checkRecord = await checkInToday(user.address.toLowerCase(), dateTag)
    let todayStat = 0
    if (checkRecord) {
      // 有签到的链上记录, 说明已签到
      todayStat = 1
    }
    if (record.lastSignDay === dateTag) {
      // 检查是否已领取
      todayStat = 2
    }
    const scoreBonus = seqSignScore(checkRecord?.count || 0)
    const ticketRecords = await TicketRecord.find({ user: user.id, activity: user.activity, type: SIGN_TOTAL })
    const claimedSet = new Set()
    ticketRecords.forEach(record => {
      claimedSet.add(record.data.day)
    })

    const totalStat = []
    for (let cfg of totalSignCfg) {
      let state = 0
      if (cfg.days <= checkRecord?.total || 0) {
        state = 1
      }
      if (claimedSet.has(cfg.days)) {
        state = 9
      }
      totalStat.push({
        days: cfg.days,
        tickets: cfg.reward,
        state,
      })
    }
    totalStat.sort((a, b) => a.days - b.days)
    return {
      ticket: record.tickets,
      signCfg,
      todayStat,
      todayTickets: 1 + scoreBonus,
      daysTotal: checkRecord?.total || 0,
      daysSeq: checkRecord?.count || 0,
      totalStat,
    }
  }
  /**
   * 探索
   */
  @router('post /api/game/step')
  async gameStep(req, res) {
    new SyncLocker().checkLock(req)
    const user = req.user
    let { step } = req.params
    step = step || '1'
    if (isNaN(step)) {
      throw new ZError(11, 'invalid step')
    }
    // check if step is safe int
    step = parseInt(step)
    if (step < 1) {
      step = 1
    }
    // const session = await mongoose.startSession()
    // session.startTransaction()
    // try {
    const record = await ActivityGame.insertOrUpdate({ user: user.id, activity: user.activity }, {})
    if (MANUAL_OPEN_GAME && record.status === 0) {
      throw new ZError(12, 'map not open')
    }
    if (record.tickets < step) {
      throw new ZError(13, 'insufficient tickets')
    }
    const ticketRecord = new TicketRecord({
      user: user.id,
      activity: user.activity,
      type: USE_TICKET,
      data: {},
      score: -step,
    })
    let updateData: any = {
      $inc: { tickets: -step },
    }
    let { score, chests } = generateStepReward(user.id, user.activity, step)
    if (chests.length > 0) {
      updateData.maxNoChestCount = 0
    } else {
      if (record.maxNoChestCount + step >= RESET_STEP) {
        updateData.maxNoChestCount = 0
        const level = generateChestLevel()
        chests.push(generateNewChest(user.id, user.activity, level, ChestStatusEnum.NORMAL))
      } else {
        updateData['$inc']['maxNoChestCount'] = step
      }
    }
    await ActivityGame.updateOne({ user: user.id, activity: user.activity }, updateData)
    await updateRankScore({
      user: user.id,
      score: score,
      activity: user.activity,
      scoreType: SCORE_GAME_STEP,
      scoreParams: {
        step: step,
      },
    })
    // batch save chests
    for (let chest of chests) {
      await chest.save()
    }
    await ticketRecord.save()
    // await ticketRecord.save({ session })
    // await session.commitTransaction()
    // session.endSession()
    return { score, chests: chests.map(chest => chest.toJson()) }
    // } catch (e) {
    //   session.abortTransaction()
    //   session.endSession()
    //   throw e
    // }
  }
}
